// Exact stochastic watersheds, C++ implementation
// Filip Malmberg 2013.				
// For description, license, and example usage see README.TXT.

#include "StochasticWS.h"
#include "Edge.h"
#include "DisjointSet.h"
#include <vector>
#include <queue>
#include <set>
#include <algorithm>
#include <math.h>

using namespace std;


Graph StochasticWS(Graph& g, int n_seeds){
 
  long n=g.GetNodeCount();

  Graph result(n);

  DisjointSet ds(n);

  vector<Edge> edges=g.GetEdgesAsVector();

  vector<long> RegionSize(n);
  
  for(long i=0; i<n; i++){
    RegionSize[i]=1;
  }

  CompareEdgeByWeightInv ComparisonObject;
  std::sort(edges.begin(), edges.end(), ComparisonObject);

  long p_index, q_index, merged_index;
  long p_size, q_size;
  double probability;

  for(long i=0; i<edges.size(); i++){
    p_index=ds.Find(edges[i].p);
    q_index=ds.Find(edges[i].q);
    
    p_size=RegionSize[p_index];
    q_size=RegionSize[q_index];
    
    //Calculate edge weight based on p_size and q_size
    probability=1-pow( 1-double(p_size)/double(n) , n_seeds) - pow( 1-double(q_size)/double(n) , n_seeds) + pow( 1-double(p_size+q_size)/double(n) , n_seeds);
    result.SetEdge(edges[i].p, edges[i].q, probability );

    merged_index=ds.Union(p_index,q_index);
    RegionSize[merged_index]=p_size+q_size;
 
  }

  return result;

}

Graph StochasticWS_naive(Graph& g, int n_seeds ){
  long n=g.GetNodeCount();
  Graph result(n);

  vector<Edge> edges=g.GetEdgesAsVector();
  
  long p_size, q_size;
  long index;

  double weight;
  double probability;

  queue<long> Q; //FIFO queue
  vector<bool> visited (n);
  map<int,double> neighbors;

  for(long i=0; i<edges.size(); i++){
    //cout<<"edge: "<<i<<endl;

    weight=edges[i].w;

    //Determine p_size by breadth-first search from p
    fill(visited.begin(), visited.end(), false);
    p_size=0;
    Q.push(edges[i].p);
    while(!Q.empty()){
      p_size++;
      index=Q.front();
      Q.pop();
      visited[index]=true;
      neighbors= g.GetAdjacent(index);
      for (auto it=neighbors.begin(); it!=neighbors.end(); it++){
	if(!visited[it->first] && it->second < weight){
	  Q.push(it->first);
	}
      }
    }

    //Determine q_size by breadth-first search from q
    fill(visited.begin(), visited.end(), false);
    q_size=0;
    Q.push(edges[i].q);
    while(!Q.empty()){
      q_size++;
      index=Q.front();
      Q.pop();
      visited[index]=true;
      neighbors= g.GetAdjacent(index);
      for (auto it=neighbors.begin(); it!=neighbors.end(); it++){
	if(!visited[it->first] && it->second < weight){
	  Q.push(it->first);
	}
      }
    }

    
    //Calculate edge weight based on p_size and q_size
    probability=1-pow( 1-double(p_size)/double(n) , n_seeds) - pow( 1-double(q_size)/double(n) , n_seeds) + pow( 1-double(p_size+q_size)/double(n) , n_seeds);
    result.SetEdge(edges[i].p, edges[i].q, probability );
    

  }
  return result;

}
